# PROBEER OOIT NOG EEN KEER OM DE countvec OP MAX TE ZETTEN OFZO
# https://github.com/paulgb/sklearn-pandas

# from sklearn.decomposition import PCA
import copy

import numpy as np
import pandas as pd
import scipy.sparse
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.feature_extraction.text import CountVectorizer
from sklearn.preprocessing import (OneHotEncoder, PolynomialFeatures,
                                   robust_scale)

from dateutil.parser import parse
from xtoy.utils import is_numeric
from xtoy.utils import is_integer


class DataFrameImputer(TransformerMixin):
    """
    Credits http://stackoverflow.com/a/25562948/1575066
    """

    def __init__(self):
        """Impute missing values.

        Columns of dtype object are imputed with the most frequent value
        in column.

        Columns of other types are imputed with mean of column.

        """
        self.fill = None

    @staticmethod
    def most_frequent(col):
        try:
            return col.value_counts().index[0]
        except IndexError:
            return 0

    def get_impute_val(self, col):
        if col.dtype == np.dtype('O'):
            val = self.most_frequent(col)
        elif col.dtype == np.dtype(float):
            val = col.mean()
        else:
            val = col.median()
        if isinstance(val, float) and np.isnan(val):
            val = 0
        return val

    def fit(self, X, y=None):
        self.fill = pd.Series([self.get_impute_val(X[c]) for c in X], index=X.columns)
        return self

    def transform(self, X, y=None):
        return X.fillna(self.fill, inplace=False)


class Sparsify(BaseEstimator, TransformerMixin):

    def __init__(self,
                 count_vectorizer=CountVectorizer(max_features=100, token_pattern=r"(?u)\b\w+\b"),
                 max_unique_for_discrete=15,
                 date_atts=("year", "month", "day", "weekday", "hour", "minute", "second")):
        self.count_vectorizer = count_vectorizer
        self.max_unique_for_discrete = max_unique_for_discrete
        self.one_hot_encoder = None
        # scary duplication
        self.drop_vars = []
        self.count_vecs = []
        self.date_vars = []
        self.imputer = DataFrameImputer()
        self.ohe_indices = []
        self.numeric_indices = []
        self.var_names_ = None
        self.date_atts = date_atts

    def fit(self, X, y=None):
        if not isinstance(X, pd.DataFrame):
            X = pd.DataFrame(X)
        X = self.imputer.fit_transform(X)
        self.ohe_indices = np.zeros(X.shape[1], dtype=bool)
        self.numeric_indices = np.zeros(X.shape[1], dtype=bool)
        self.drop_vars = np.zeros(X.shape[1], dtype=bool)
        self.date_vars = np.zeros(X.shape[1], dtype=bool)
        self.var_names_ = []
        for i, col in enumerate(X):
            if len(set(X[col])) == 1:
                self.drop_vars[i] = True

        for i, col in enumerate(X):
            if self.drop_vars[i]:
                continue
            if is_numeric(X[col]):
                if len(set(X[col])) == 2:
                    self.var_names_.append('{}_{}_continuous'.format(col, i))
                else:
                    self.var_names_.append('{}_{}_dummy'.format(col, i))
                num_unique = len(np.unique(X[col]))
                if is_integer(X[col]) and 3 <= num_unique <= self.max_unique_for_discrete:
                    self.ohe_indices[i] = True
                self.numeric_indices[i] = True
            # maybe date
            else:
                for x in X[col]:
                    try:
                        parse(x)
                    except ValueError:
                        break
                else:
                    self.date_vars[i] = True
                    date_var_names = ['{}_date_{}'.format(col, x)
                                      for j, x in enumerate(self.date_atts)]
                    self.var_names_.extend(date_var_names)

        for i, (ohed, col) in enumerate(zip(self.ohe_indices, X)):
            if not self.drop_vars[i] and not is_numeric(X[col]) and not ohed and not self.date_vars[i]:
                countvec = copy.copy(self.count_vectorizer)
                countvec.fit(['' if isinstance(x, float) else x for x in X[col]])
                self.count_vecs.append(countvec)
                cv_var_names = ['{}_countvec_{}_{}_{}'.format(col, x, i, j)
                                for j, x in enumerate(countvec.get_feature_names())]
                self.var_names_.extend(cv_var_names)
            else:
                self.count_vecs.append(False)

        self.one_hot_encoder = OneHotEncoder(handle_unknown='ignore')
        if np.any(self.ohe_indices):
            self.one_hot_encoder.fit(X[X.columns[self.ohe_indices]])
            # ridiculously complex OHE feature range names
            feat = set(self.one_hot_encoder.active_features_)
            for l, h, v in zip(self.one_hot_encoder.feature_indices_[:-1],
                               self.one_hot_encoder.feature_indices_[1:],
                               X.columns[self.ohe_indices]):
                self.var_names_.extend(['{}_OHE_{}'.format(v, j)
                                        for j in range(len(feat.intersection(range(l, h))))])
        return self

    def transform(self, X, y=None):
        if not isinstance(X, pd.DataFrame):
            X = pd.DataFrame(X)
        X = self.imputer.transform(X)
        res = []
        cvs = []
        date_groups = []
        for i, col in enumerate(X):
            if self.drop_vars[i] or self.ohe_indices[i] or self.count_vecs[i]:
                continue
            if self.numeric_indices[i]:
                res.append(X[col])
            elif self.date_vars[i]:
                prefix = str(col) + "__"
                dtimes = [parse(x) for x in X[col]]
                for a in self.date_atts:
                    dcol = [x.weekday() if a == "weekday" else getattr(x, a) for x in dtimes]
                    res.append(dcol)
        for cv, col in zip(self.count_vecs, X):
            if cv:
                cvs.append(cv.transform(['' if isinstance(x, (np.int64, int, float)) else x
                                         for x in X[col]]))
        combined = [scipy.sparse.coo_matrix(res).T] + cvs
        if np.any(self.ohe_indices):
            ohd = self.one_hot_encoder.transform(X[X.columns[self.ohe_indices]])
            combined = combined + [ohd]

        return scipy.sparse.hstack(combined).tocsr()



#from sklearn.svm import SVC
import scipy
from sklearn import svm
from sklearn.decomposition import TruncatedSVD
from sklearn.ensemble import RandomForestClassifier, RandomForestRegressor
from sklearn.linear_model import Ridge, RidgeClassifier
from sklearn.neighbors import KNeighborsClassifier

from xtoy.utils import classification_or_regression


ridge_grid = {'clf__alpha': [0.00001, 0.001, 0.01, 0.05, 0.1, 0.3, 0.5,
                             0.8, 1.0, 1.4, 1.7, 10, 100, 1000, 1000000]}
# normalize seems bugged at prediction time!

#ridge_grid = {'clf__alpha': [0.1, 1., 10.]}

ridge_classification = {'clf': RidgeClassifier, 'grid': ridge_grid}
ridge_regression = {'clf': Ridge, 'grid': ridge_grid}
# ridge_regression = {'clf': svm.SVR, 'grid': {
#     'clf__shrinking': [True, False],
#     'clf__C': [0.001, 0.01, 0.1, 1, 10, 100, 1000],
#     'clf__kernel': ['linear', 'poly', 'rbf', 'sigmoid']}}


rf_grid = {'clf__max_features': ['sqrt', 'auto', 'log2', 0.5, 0.8, 0.9],
           'clf__max_depth': [None, 5, 10, 20],
           'clf__min_samples_leaf': [1, 5, 10],
           'clf__min_samples_split': [1, 10, 20],
           'clf__n_estimators': [50]}

rf_grid_classification = rf_grid.copy()
rf_grid_classification.update({'clf__class_weight': ['balanced']})

rf_classification = {'clf': RandomForestClassifier, 'grid': rf_grid_classification}
rf_regression = {'clf': RandomForestRegressor, 'grid': rf_grid}

# ridge_classification = {'clf': KNeighborsClassifier, 'grid': {
#    "clf__n_neighbors": [4], "clf__weights": ["distance"]}}


options = {
    'regression': {
        'dense': [ridge_regression, rf_regression],
        'sparse': [ridge_regression]
    },
    'classification': {
        'dense': [ridge_classification, rf_classification],
        'sparse': [ridge_classification]
    }
}


def sparse_or_dense(X, RAM=None, magic=42):
    if scipy.sparse.issparse(X):
        return "sparse"
    else:
        # USING PAPER WE WILL GO TO DENSE ANYWAY
        return 'dense'
    # size = np.prod(X.shape) if hasattr(X, 'shape') else len(X) * len(X[0])

    # # if N * num_feat * magic > RAM:
    # if size > 1000 ** 3:
    #     return 'sparse'
    # else:
    #     return 'dense'


def pick(X, y, cl_or_reg=None, opts=None):
    # i always choose first one now
    if opts is None:
        opts = options
    cl_or_reg = cl_or_reg if cl_or_reg else classification_or_regression(y)
    return opts[cl_or_reg][sparse_or_dense(X)][0]



import numpy as np
import pandas as pd
from sklearn.datasets import load_boston
from sklearn.datasets import load_breast_cancer
from sklearn.datasets import load_diabetes
from sklearn.datasets import load_digits
from sklearn.datasets import load_iris
from sklearn.datasets import load_linnerud
from sklearn.datasets import fetch_20newsgroups
from sklearn.datasets import make_regression

from xtoy.toys import Toy


def apply_toy_on(X, y, cl_or_reg=None):
    if not isinstance(X, pd.DataFrame):
        X = pd.DataFrame(X)
    if not isinstance(y, pd.DataFrame):
        y = pd.DataFrame(y)
    toy = Toy(cl_or_reg=cl_or_reg)
    toy.fit(X[0::2], y[0::2])
    return toy.score(X[1::2], y[1::2])


# def test_date():
#     X = [["2015-04"], ["2015-05"], ["2015-06"]] * 10
#     y = [1, 1, 0] * 10
#     assert apply_toy_on(X, y) > 0.5
from xtoy import Sparsify


def test_digits_data():
    X, y = load_digits(return_X_y=True)
    assert apply_toy_on(X, y) > 0.7


def test_iris_data():
    X, y = load_iris(return_X_y=True)
    assert apply_toy_on(X, y) > 0.1


def test_boston_data():
    X, y = load_boston(return_X_y=True)
    assert apply_toy_on(X, y)


def test_breast_cancer_data():
    X, y = load_breast_cancer(return_X_y=True)
    assert apply_toy_on(X, y) > 0.3


def test_diabetes_data():
    X, y = load_diabetes(return_X_y=True)
    assert apply_toy_on(X, y)

# problem with really small data, maybe multiply cases like that


def test_missing():
    X = np.array([1, 2, 3, 4, np.nan, 5, np.nan, np.nan, np.nan, np.nan] * 10)
    y = np.array([0, 0, 0, 0, 0, 1, 1, 1, 1, 1] * 10)
    assert apply_toy_on(X, y) > 0.1


def test_text_missing():
    X = np.array(["1", "2", "3", "4", None, None, "5", None, None, None, None] * 10)
    y = np.array([0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1] * 10)
    assert apply_toy_on(X, y) > 0.1

# # sensitive to splitting technique
# def test_missing():
#     X = np.array([1, 2, 3, 4, 5, np.nan, np.nan, np.nan, np.nan, np.nan])
#     y = np.array([0, 0, 0, 0, 0, 1, 1, 1, 1, 1])
#     assert apply_toy_on(X, y) > 0.1


# def test_text_missing():
#     X = np.array(["1", "2", "3", "4", "5", None, None, None, None, None])
#     y = np.array([0, 0, 0, 0, 0, 1, 1, 1, 1, 1])
#     assert apply_toy_on(X, y) > 0.1


# multi output regression
def test_linnerud_data():
    X, y = load_linnerud(return_X_y=True)
    assert apply_toy_on(X, y) > -3000


def test_date_missing_data():
    data = pd.read_csv("../data/nba_2016.csv")
    X, y = data.drop("PTS.1", axis=1), data["PTS.1"]
    assert apply_toy_on(X, y)


def test_make_reg():
    X, y = make_regression(1000)
    assert apply_toy_on(X, y)


def test_newsgroup_data():
    categories = ['alt.atheism', 'talk.religion.misc', 'comp.graphics', 'sci.space']

    newsgroups = fetch_20newsgroups(subset='train',
                                    remove=('headers', 'footers', 'quotes'),
                                    categories=categories)

    X, y = pd.DataFrame(newsgroups.data), newsgroups.target
    assert apply_toy_on(X, y, cl_or_reg='classification')



import pandas as pd
import numpy as np
from sklearn.pipeline import Pipeline
from sklearn.preprocessing import Normalizer

from xtoy.evolutionary_search import EvolutionaryAlgorithmSearchCV as evo_search
from xtoy.prep import Sparsify
from xtoy.classifiers import pick
from xtoy.classifiers import classification_or_regression

from xtoy.scorers import f1_weighted_scorer
from xtoy.scorers import mse_scorer

from xtoy.utils import get_cv_splits

try:
    import pickle
except (ValueError, SystemError, ImportError):
    pass


class Toy:

    def __init__(self, cv=get_cv_splits, scoring=None, n_jobs=1, cl_or_reg=None, **kwargs):
        self.cv = get_cv_splits
        self.clf = None
        self.pipeline = None
        self.evo = None
        self.scoring = scoring
        self.n_jobs = n_jobs
        self.kwargs = kwargs
        self.grid = None
        self.cl_or_reg = cl_or_reg
        self.sparsify = Sparsify()

    def pick_model(self, X, y):
        chosen_model = pick(X, y, self.cl_or_reg)
        self.clf = chosen_model['clf']
        self.grid = chosen_model['grid']

    def get_pipeline(self):
        self.pipeline = Pipeline(steps=[
            #('tsvd', TruncatedSVD()),  # this one also has to have % top features chosen
            #('feature_selection', SelectFromModel(Ridge())),
            ("scaler", Normalizer()),
            ('clf', self.clf())
        ])

    def fit(self, X, y):
        if not isinstance(X, pd.DataFrame):
            X = pd.DataFrame(X)
        X = pd.DataFrame(self.sparsify.fit_transform(X).A)
        if self.scoring is None:
            tp = classification_or_regression(y)
            self.scoring = [f1_weighted_scorer, mse_scorer][tp != 'classification']
        print(self.scoring)
        self.pick_model(X, y)
        self.get_pipeline()
        unique_combinations = np.prod(list(map(len, self.grid.values())))
        print("unique_combinations", unique_combinations)
        if 'population_size' not in self.kwargs:
            self.kwargs['population_size'] = np.clip(int(unique_combinations / 1000), 5, 10)
        if 'generations_number' not in self.kwargs:
            self.kwargs['generations_number'] = np.clip(int(unique_combinations / 20), 10, 50)

        self.evo = evo_search(self.pipeline, self.grid, scoring=self.scoring,
                              cv=self.cv, n_jobs=self.n_jobs, **self.kwargs)
        self.evo.fit(X, y)
        return self.evo.best_estimator_

    def predict(self, X):
        X = self.sparsify.transform(X).A
        return self.evo.predict(X)

    def score(self, X, y):
        X = self.sparsify.transform(X).A
        return self.evo.best_estimator_.score(X, y)

    # def baselines():
    #     f1_weighted_score

    def best_model_pickle(self):
        return pickle.dumps(self.evo.best_estimator_)



something is an ID when:
- multisets, when there is high value overlap with another key OR same name of var
- low linear correlation(comparatively?)
- first, or last var index?
- all unique values, or groups

something is continuous when:
- numeric, more than 10 values, has some linear correlation?

something is discrete when:
- numeric, < 10 values, has some linear correlation

something is categorical when:
- string, number of unique values < 100
- numeric, number of unique values < 10, no correlation(difference with discrete is if we want to sparse it)

something is corpus when:
- string, after CountVect it has overlap contains spaces probably

something is a date when:
- can be parsed as date, either - , / or ' '
- year:
    max 100 unique values, all between 1900 - 2100
- month:
    1 - 12 only, and year found
- day:
    1 - 31, and month or year found


numeric, ID                    converts to        empty
numeric, continuous            converts to        scale(float(1)), sparse(ohe(bins(m)))
numeric, discrete              converts to        scale(float(1)), sparse(ohe(m))
numeric, categorical           converts to        ohe(m), sparse(ohe(m))
numeric, date                  converts to
string, categorical            converts to        ohe(m), sparse(ohe(m))
string, corpus                 converts to
string, ID                     converts to
date                           converts to        float(4), sparse(Y, M, D, S)


numeric, continuous, nomiss       impute, scale
numeric, continuous, miss
numeric, discrete, nomiss
numeric, discrete, miss
numeric, date, nomiss
numeric, date, miss
string, categorical
string, corpus

import scipy.sparse

scipy.sparse.csr_matrix(X)

decide if ID, 'probably no ID, otherwise most likely: <>', 'most likely <>'


def is_sorted(l):
    return all(l[i] <= l[i + 1] for i in range(len(l) - 1))


def is_float(x)
    try:
        if int(x) != float(x):
            return True
    except ValueError:
        pass
    return False


def find_id_in_df(m, columns=None):
    ids = []
    mscore = 0
    for i, c in enumerate(m.T):
        c = c[:10000]
        if is_float(c[0]):
            continue
        score = 0
        if columns is not None:
            if 'id' in columns[i].lower():
                score += 1
        if is_sorted(c):
            score += 1
        # even if 'nan' it wouldn't be likely to be id
        if len(c) == len(set(c)):
            score += 1
        ids.append((i, score))
        mscore = max(mscore, score)
    if mscore > 0:
        return [i[0] for i in ids if mscore == i[1]]
    return None


#

(dense, string, categorical):
string, categorical -> DictVectorizer(sparse=False)
(sparse, string, categorical):
string, categorical -> DictVectorizer(sparse=False)

sniffer.consider_max_values

from collections import namedtuple

allowed_subclasses = namedtuple('AllowedSubclasses',
                                'id continuous discrete categorical date corpus')
allowed_subclasses.__new__.__defaults__ = (None, None, None, None, None, None)




from sklearn.metrics import mean_squared_error
from sklearn.metrics import f1_score
from sklearn.metrics import make_scorer


def f1_weighted_score(*args, **kwargs):
    return f1_score(average="weighted", *args, **kwargs)


def mse(*args, **kwargs):
    return -mean_squared_error(*args, **kwargs)


f1_weighted_scorer = make_scorer(f1_weighted_score)
mse_scorer = make_scorer(mse)



from setuptools import setup
from setuptools import find_packages

MAJOR_VERSION = '0'
MINOR_VERSION = '2'
MICRO_VERSION = '39'
VERSION = "{}.{}.{}".format(MAJOR_VERSION, MINOR_VERSION, MICRO_VERSION)

setup(name='xtoy',
      version=VERSION,
      description='get xtoyed predictions from raw data',
      url='https://github.com/kootenpv/xtoy',
      author='Pascal van Kooten',
      author_email='kootenpv@gmail.com',
      license='MIT',
      packages=find_packages(),
      install_requires=[
          'numpy',
          'scipy',
          'pandas',
          'scikit-learn>=0.18.1',
          'deap',
          'bitstring'
      ],
      classifiers=[
          'Environment :: Console',
          'Intended Audience :: Developers',
          'Intended Audience :: Science/Research',
          'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
          'Operating System :: Microsoft',
          'Operating System :: MacOS :: MacOS X',
          'Operating System :: Unix',
          'Operating System :: POSIX',
          'Programming Language :: Python',
          'Programming Language :: Python :: 2.7',
          'Programming Language :: Python :: 3',
          'Programming Language :: Python :: 3.3',
          'Programming Language :: Python :: 3.4',
          'Programming Language :: Python :: 3.5',
          'Topic :: Scientific/Engineering :: Artificial Intelligence',
          'Topic :: Software Development',
          'Topic :: Utilities'
      ],
      keywords=['data science', 'machine learning', 'genetic programming'],
      zip_safe=False,
      platforms='any')
